/****************************************************************************
*                      SPLAT: A Terrain Analysis Program                    *
*               Copyright John A. Magliacane, KD2BD 1997-2002               *
*                         Last update: 10-Jun-2002                          *
*****************************************************************************
*                                                                           *
* This program is free software; you can redistribute it and/or modify it   *
* under the terms of the GNU General Public License as published by the     *
* Free Software Foundation; either version 2 of the License or any later    *
* version.                                                                  *
*                                                                           *
* This program is distributed in the hope that it will useful, but WITHOUT  *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     *
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License     *
* for more details.                                                         *
*                                                                           *
*****************************************************************************
*         To compile: gcc -Wall -O3 -s -lm -lbz2 splat.c -o splat           *
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <bzlib.h>
#include "fontdata.h"

#define GAMMA 2.5
#define MAXSLOTS 9
#define BZBUFFER 65536

#if MAXSLOTS==4
#define ARRAYSIZE 4950
#endif

#if MAXSLOTS==9
#define ARRAYSIZE 10870
#endif

#if MAXSLOTS==16
#define ARRAYSIZE 19240
#endif

#if MAXSLOTS==25
#define ARRAYSIZE 30025
#endif

char string[255], sdf_path[255], *splat_version={"1.0.3"};

double	TWOPI=6.283185307179586, HALFPI=1.570796326794896,
	PI=3.141592653589793, deg2rad=1.74532925199e-02,
	EARTHRADIUS=20902230.97;

int	min_north=0, max_north=0, min_west=0, max_west=0,
	max_elevation=0, min_elevation=0, bzerror;

struct site { double lat;
              double lon;
              double alt;
              char name[50];
            };

struct	{ float lat[ARRAYSIZE];
	  float lon[ARRAYSIZE];
	  float elevation[ARRAYSIZE];
	  float distance[ARRAYSIZE];
	  int length;
	} path;

struct { int min_north;
         int max_north;
         int min_west;
         int max_west;
         int max_el;
         int min_el;
         int data[1200][1200];
	 unsigned char mask[1200][1200];
       } dem[MAXSLOTS];

double arccos(x,y)
double x, y;
{
	/* This function implements the arc cosine function,
	   returning a value between 0 and TWOPI. */

	double result=0.0;

	if (y>0.0)
		result=acos(x/y);

	if (y<0.0)
		result=PI+acos(x/y);

	return result;
}

char *dec2dms(decimal)
double decimal;
{
	/* Converts decimal degrees to degrees, minutes, seconds,
	   (DMS) and returns the result as a character string. */
 
	int degrees, minutes, seconds;
	double a, b, c, d;

	a=floor(decimal);
	b=60.0*(decimal-a);
	c=floor(b);
	d=60.0*(b-c);

	degrees=(int)a;
	minutes=(int)c;
	seconds=(int)rint(d);

	string[0]=0;
	sprintf(string,"%d%c %d\' %d\"", degrees, 176, minutes, seconds);
	return (string);
}

int OrMask(lat,lon,value)
double lat, lon;
int value;
{
	/* Lines, text, markings, and coverage areas are stored in a
	   mask that is combined with topology data when topographic
	   maps are generated by SPLAT!.  This function sets bits in
	   the mask based on the latitude and longitude of the area
	   pointed to. */

	int x, y, indx, minlat, minlon;
	char found;

	minlat=(int)floor(lat);
	minlon=(int)floor(lon);

	for (indx=0, found=0; indx<MAXSLOTS && found==0;)
		if (minlat==dem[indx].min_north && minlon==dem[indx].min_west)
			found=1;
		else
			indx++;

	if (found)
	{
		x=(int)(1199.0*(lat-floor(lat)));
		y=(int)(1199.0*(lon-floor(lon)));

		dem[indx].mask[x][y]|=value;
		return (dem[indx].mask[x][y]);
	}
	else
		return -1;
}

int GetMask(lat,lon)
double lat, lon;
{
	/* This function returns the mask bits based on the latitude
	   and longitude given. */

	return (OrMask(lat,lon,0));
}

double GetElevation(location)
struct site location;
{
	/* This function returns the elevation (in feet) of any location
	   represented by the digital elevation model data in memory.
	   Function returns -5000.0 for locations not found in memory. */

	char found;
	int x, y, indx, minlat, minlon;
	double elevation;

	elevation=-5000.0;

	minlat=(int)floor(location.lat);
	minlon=(int)floor(location.lon);

	x=(int)(1199.0*(location.lat-floor(location.lat)));
	y=(int)(1199.0*(location.lon-floor(location.lon)));

	for (indx=0, found=0; indx<MAXSLOTS && found==0; indx++)
	{
		if (minlat==dem[indx].min_north && minlon==dem[indx].min_west)
		{
			elevation=3.28084*dem[indx].data[x][y];
			found=1;
		}
	}
	
	return elevation;
}

double Distance(site1, site2)
struct site site1, site2;
{
	/* This function returns the great circle distance
	   in miles between any two site locations. */

	double lat1, lon1, lat2, lon2, distance;

	lat1=site1.lat*deg2rad;
	lon1=site1.lon*deg2rad;
	lat2=site2.lat*deg2rad;
	lon2=site2.lon*deg2rad;

	distance=3959.0*acos(sin(lat1)*sin(lat2)+cos(lat1)*cos(lat2)*cos((lon1)-(lon2)));

	return distance;
}

double Azimuth(source, destination)
struct site source, destination;
{
	/* This function returns the azimuth (in degrees) to the
	   destination as seen from the location of the source. */

	double dest_lat, dest_lon, src_lat, src_lon,
	       beta, azimuth, diff, num, den, fraction;

	dest_lat=destination.lat*deg2rad;
	dest_lon=destination.lon*deg2rad;

	src_lat=source.lat*deg2rad;
	src_lon=source.lon*deg2rad;
		
	/* Calculate Surface Distance */

	beta=acos(sin(src_lat)*sin(dest_lat)+cos(src_lat)*cos(dest_lat)*cos(src_lon-dest_lon));

	/* Calculate Azimuth */

	num=sin(dest_lat)-(sin(src_lat)*cos(beta));
	den=cos(src_lat)*sin(beta);
	fraction=num/den;

	/* Trap potential problems in acos() due to rounding */

	if (fraction>=1.0)
		fraction=1.0;

	if (fraction<=-1.0)
		fraction=-1.0;

	/* Calculate azimuth */

	azimuth=acos(fraction);

	/* Reference it to True North */

	diff=dest_lon-src_lon;

	if (diff<=-PI)
		diff+=TWOPI;

	if (diff>=PI)
		diff-=TWOPI;

	if (diff>0.0)
		azimuth=TWOPI-azimuth;

	return (azimuth/deg2rad);		
}

double ElevationAngle(local, remote)
struct site local, remote;
{
	/* This function returns the angle of elevation (in degrees)
	   of the remote location as seen from the local site.
	   A positive result represents an angle of elevation (uptilt),
	   while a negative result represents an angle of depression
	   (downtilt), as referenced to a normal to the center of
	   the earth. */
	   
	register double a, b, dx;

	a=GetElevation(remote)+remote.alt+EARTHRADIUS;
	b=GetElevation(local)+local.alt+EARTHRADIUS;
	dx=5280.0*Distance(local,remote);

	/* Apply the Law of Cosines */

	return ((180.0*(acos(((b*b)+(dx*dx)-(a*a))/(2.0*b*dx)))/PI)-90.0);
}

void ReadPath(source, destination)
struct site source, destination;
{
	/* This function generates a sequence of latitude and
	   longitude positions between a source location and
	   a destination along a great circle path, and stores
	   elevation and distance information for points along
	   that path in the "path" structure for later use. */

	double azimuth, distance, lat1, lon1, beta,
	       den, num, lat2, lon2, total_distance;
	int x0, y0, x1, y1, c;
	struct site tempsite;

	c=0;
	x0=0;
	y0=0;

	lat1=source.lat*deg2rad;
	lon1=source.lon*deg2rad;
	azimuth=Azimuth(source,destination)*deg2rad;
	total_distance=Distance(source,destination);

	for (distance=0; distance<=total_distance; distance+=0.04)
	{
		beta=distance/3959.0;
		lat2=asin(sin(lat1)*cos(beta)+cos(azimuth)*sin(beta)*cos(lat1));
		num=cos(beta)-(sin(lat1)*sin(lat2));
		den=cos(lat1)*cos(lat2);

		if (azimuth==0.0 && (beta>HALFPI-lat1))
			lon2=lon1+PI;

		else if (azimuth==HALFPI && (beta>HALFPI+lat1))
			lon2=lon1+PI;

		else if (fabs(num/den)>1.0)
			lon2=lon1;

		else
		{
			if ((PI-azimuth)>=0.0)
				lon2=lon1-arccos(num,den);
			else
				lon2=lon1+arccos(num,den);
		}
	
		while (lon2<0.0)
			lon2+=TWOPI;

		while (lon2>TWOPI)
			lon2-=TWOPI;
 
		lat2=lat2/deg2rad;
		lon2=lon2/deg2rad;

		x1=(int)(1199.0*(lat2-floor(lat2)));
		y1=(int)(1199.0*(lon2-floor(lon2)));

		if (x1!=x0 || y1!=y0)
		{
			path.lat[c]=lat2;
			path.lon[c]=lon2;
			tempsite.lat=lat2;
			tempsite.lon=lon2;
			path.elevation[c]=GetElevation(tempsite);
			path.distance[c]=distance;
			c++;
		}

		x0=x1;
		y0=y1;
	}

	/* Make sure exact destination point is recorded at path.length-1 */

	x1=(int)(1199.0*(destination.lat-floor(destination.lat)));
	y1=(int)(1199.0*(destination.lon-floor(destination.lon)));

	path.lat[c]=destination.lat;
	path.lon[c]=destination.lon;
	path.elevation[c]=GetElevation(destination);
	path.distance[c]=total_distance;
	c++;

	path.length=c;
}

double AverageTerrain(source, azimuth, start_distance, end_distance)
struct site source;
double azimuth, start_distance, end_distance;
{
	/* This function returns the average terrain calculated in
	   the direction of "azimuth" (degrees) between "start_distance"
	   and "end_distance" (miles) from the source location.  If
	   the terrain is all water (non-critical error), -5000.0 is
	   returned.  If not enough SDF data has been loaded into
	   memory to complete the survey (critical error), then
	   -9999.0 is returned. */
 
	int c, samples, endpoint;
	double beta, lat1, lon1, lat2, lon2, num, den, terrain=0.0;
	struct site destination;

	lat1=source.lat*deg2rad;
	lon1=source.lon*deg2rad;

	/* Generate a path of elevations between the source
	   location and the remote location provided. */

	beta=end_distance/3959.0;

	azimuth=deg2rad*azimuth;

	lat2=asin(sin(lat1)*cos(beta)+cos(azimuth)*sin(beta)*cos(lat1));
	num=cos(beta)-(sin(lat1)*sin(lat2));
	den=cos(lat1)*cos(lat2);

	if (azimuth==0.0 && (beta>HALFPI-lat1))
		lon2=lon1+PI;

	else if (azimuth==HALFPI && (beta>HALFPI+lat1))
		lon2=lon1+PI;

	else if (fabs(num/den)>1.0)
		lon2=lon1;

	else
	{
		if ((PI-azimuth)>=0.0)
			lon2=lon1-arccos(num,den);
		else
			lon2=lon1+arccos(num,den);
	}
	
	while (lon2<0.0)
		lon2+=TWOPI;

	while (lon2>TWOPI)
		lon2-=TWOPI;
 
	lat2=lat2/deg2rad;
	lon2=lon2/deg2rad;

	destination.lat=lat2;
	destination.lon=lon2;

	/* If SDF data is missing for the endpoint of
	   the radial, then the average terrain cannot
	   be accurately calculated.  Return -9999.0 */

	if (GetElevation(destination)<-4999.0)
		return (-9999.0);
	else
	{
		ReadPath(source,destination);

		endpoint=path.length;

		/* Shrink the length of the radial if the
		   outermost portion is not over U.S. land. */

		for (c=endpoint-1; c>=0 && path.elevation[c]==0.0; c--);

		endpoint=c+1;

		for (c=0, samples=0; c<endpoint; c++)
		{
			if (path.distance[c]>=start_distance)
			{
				terrain+=path.elevation[c];
				samples++;
			}
		}

		if (samples==0)
			terrain=-5000.0;  /* No land */
		else
			terrain=(terrain/(double)samples);

		return terrain;
	}
}

double haat(antenna)
struct site antenna;
{
	/* This function returns the antenna's Height Above Average
	   Terrain (HAAT) based on FCC Part 73.313(d).  If a critical
	   error occurs, such as a lack of SDF data to complete the
	   survey, -5000.0 is returned. */

	int azi, c;
	char error=0;
	double terrain, avg_terrain, haat, sum=0.0;

	/* Calculate the average terrain between 2 and 10 miles
	   from the antenna site at azimuths of 0, 45, 90, 135,
	   180, 225, 270, and 315 degrees. */

	for (c=0, azi=0; azi<=315 && error==0; azi+=45)
	{
		terrain=AverageTerrain(antenna, (double)azi, 2.0, 10.0);

		if (terrain<-9998.0)  /* SDF data is missing */
			error=1;

		if (terrain>-4999.0)  /* It's land, not water */
		{
			sum+=terrain;  /* Sum of averages */
			c++;
		}
	}

	if (error)
		return -5000.0;
	else
	{
		avg_terrain=(sum/(double)c);
		haat=(antenna.alt+GetElevation(antenna))-avg_terrain;
		return haat;
	}
}

void PlaceMarker(location)
struct site location;
{
	/* This function places text and marker data in the mask array
	   for illustration on topographic maps generated by SPLAT!.
	   By default, SPLAT! centers text information BELOW the marker,
	   but may move it above, to the left, or to the right of the
	   marker depending on how much room is available on the map,
	   or depending on whether the area is already occupied by
	   another marker or label.  If no room or clear space is
	   available on the map to place the marker and its associated
	   text, then the marker and text are not are not written to
	   the map. */
 
	int	a, b, c, byte;
	char	ok2print, occupied;
	double	x, y, lat, lon, textx=0.0, texty=0.0, xmin, xmax,
		ymin, ymax, p1, p3, p6, p8, p12, p16, p24, label_length;

	xmin=min_north;
	xmax=max_north;
	ymin=min_west;
	ymax=max_west;
	lat=location.lat;
	lon=location.lon;

	if (lat<xmax && lat>xmin && lon<ymax && lon>ymin)
	{
		p1=1.0/1200.0;
		p3=3.0/1200.0;
		p6=6.0/1200.0;
		p8=8.0/1200.0;
		p12=12.0/1200.0;
		p16=16.0/1200.0;
		p24=24.0/1200.0;
		ok2print=0;
		occupied=0;

		/* Is Marker Position Clear Of Text Or Other Markers? */

		for (x=lat-p3; (x<=xmax && x>=xmin && x<=lat+p3); x+=p1)
			for (y=lon-p3; (y<=ymax && y>=ymin && y<=lon+p3); y+=p1)
				occupied|=(GetMask(x,y)&2);

		if (occupied==0)
		{
			/* Determine Where Text Can Be Positioned */

			/* label_length=length in pixels.
			   Each character is 8 pixels wide. */

			label_length=p1*(double)(strlen(location.name)<<3);

			if (((lon+label_length)<=ymax) && (lon-label_length)>=ymin)
			{
				/* Default: Centered Text */

				texty=lon+label_length/2.0;

				if ((lat-p8)>=p16)
				{
					/* Position Text Below The Marker */

					textx=lat-p8;

					x=textx;
					y=texty;
	
					/* Is This Position Clear Of
					   Text Or Other Markers? */

					for (a=0, occupied=0; a<16; a++)
					{
						for (b=0; b<strlen(location.name); b++)
							for (c=0; c<8; c++, y-=p1)
								occupied|=(GetMask(x,y)&2);
						x-=p1;
						y=texty;
					}

					x=textx;
					y=texty;

					if (occupied==0)
						ok2print=1;
				}

				else
				{
					/* Position Text Above The Marker */

					textx=lat+p24;

					x=textx;
					y=texty;
	
					/* Is This Position Clear Of
					   Text Or Other Markers? */

					for (a=0, occupied=0; a<16; a++)
					{
						for (b=0; b<strlen(location.name); b++)
							for (c=0; c<8; c++, y-=p1)
								occupied|=(GetMask(x,y)&2);
						x-=p1;
						y=texty;
					}

					x=textx;
					y=texty;

					if (occupied==0)
						ok2print=1;
				}
			}

			if (ok2print==0)
			{
				if ((lon-label_length)>=ymin)
				{
					/* Position Text To The
					   Right Of The Marker */

					textx=lat+p6;
					texty=lon-p12;

					x=textx;
					y=texty;
	
					/* Is This Position Clear Of
					   Text Or Other Markers? */

					for (a=0, occupied=0; a<16; a++)
					{
						for (b=0; b<strlen(location.name); b++)
							for (c=0; c<8; c++, y-=p1)
								occupied|=(GetMask(x,y)&2);
						x-=p1;
						y=texty;
					}

					x=textx;
					y=texty;

					if (occupied==0)
						ok2print=1;
				}

				else
				{
					/* Position Text To The
					   Left Of The Marker */

					textx=lat+p6;
					texty=lon+p8+(label_length);

					x=textx;
					y=texty;
	
					/* Is This Position Clear Of
					   Text Or Other Markers? */

					for (a=0, occupied=0; a<16; a++)
					{
						for (b=0; b<strlen(location.name); b++)
							for (c=0; c<8; c++, y-=p1)
								occupied|=(GetMask(x,y)&2);
						x-=p1;
						y=texty;
					}

					x=textx;
					y=texty;

					if (occupied==0)
						ok2print=1;
				}
			}

			/* textx and texty contain the latitude and longitude
			   coordinates that describe the placement of the text
			   on the map. */
	
			if (ok2print && textx!=0.0 && texty!=0.0)
			{
				/* Draw Text */

				x=textx;
				y=texty;

				for (a=0; a<16 && ok2print; a++)
				{
					for (b=0; b<strlen(location.name); b++)
					{
						byte=fontdata[16*(location.name[b])+a];

						for (c=128; c>0; c=c>>1, y-=p1)
							if (byte&c)
								OrMask(x,y,2);
					}
					x-=p1;
					y=texty;
				}
	
				/* Draw Square Marker Centered
				   On Location Specified */
	
				for (x=lat-p3; (x<=xmax && x>=xmin && x<=lat+p3); x+=p1)
					for (y=lon-p3; (y<=ymax && y>=ymin && y<=lon+p3); y+=p1)
						OrMask(x,y,2);
			}
		}
	}
}

double ReadBearing(input)
char *input;
{
	/* This function takes numeric input in the form of a character
	   string, and returns an equivalent bearing in degrees as a
	   decimal number (double).  The input may either be expressed
	   in decimal format (40.139722) or degree, minute, second
	   format (40 08 23).  This function also safely handles
	   extra spaces found either leading, trailing, or
	   embedded within the numbers expressed in the
	   input string.  Decimal seconds are permitted. */
 
	double bearing=0.0;
	char string[20];
	int a, b, length, degrees, minutes;
	double seconds;

	/* Copy "input" to "string", and ignore any extra
	   spaces that might be present in the process. */

	string[0]=0;
	length=strlen(input);

	for (a=0, b=0; a<length && a<18; a++)
	{
		if ((input[a]!=32 && input[a]!='\n') || (input[a]==32 && input[a+1]!=32 && b!=0))
		{
			string[b]=input[a];
			b++;
		}	 
	}

	string[b]=0;

	/* Count number of spaces in the clean string. */

	length=strlen(string);

	for (a=0, b=0; a<length; a++)
		if (string[a]==32)
			b++;

	if (b==0)  /* Decimal Format (40.139722) */
		sscanf(string,"%lf",&bearing);

	if (b==2)  /* Degree, Minute, Second Format (40 08 23) */
	{
		sscanf(string,"%d %d %lf",&degrees, &minutes, &seconds);
		bearing=(double)degrees+((double)minutes/60)+(seconds/3600);
	}

	/* Anything else returns a 0.0 */

	if (bearing>360.0 || bearing<0.0)
		bearing=0.0;

	return bearing;
}

struct site LoadQTH(filename)
char *filename;
{
	/* This function reads SPLAT! .qth (site location) files.
	   The latitude and longitude may be expressed either in
	   decimal degrees, or in degree, minute, second format.
	   Antenna height is assumed to be expressed in feet above
	   ground level (AGL), unless followed by the letter 'M',
	   or 'm', or by the word "meters" or "Meters", in which
	   case meters is assumed, and is handled accordingly. */

	int x;
	char string[50], qthfile[255];
	struct site tempsite;
	FILE *fd=NULL;

	for (x=0; filename[x]!='.' && filename[x]!=0 && x<250; x++)
		qthfile[x]=filename[x];

	qthfile[x]='.';
	qthfile[x+1]='q';
	qthfile[x+2]='t';
	qthfile[x+3]='h';
	qthfile[x+4]=0;

	tempsite.lat=0.0;
	tempsite.lon=0.0;
	tempsite.alt=0.0;
	tempsite.name[0]=0;

	fd=fopen(qthfile,"r");

	if (fd!=NULL)
	{
		/* Site Name */
		fgets(string,49,fd);

		/* Strip <CR> and/or <LF> from end of site name */

		for (x=0; string[x]!=13 && string[x]!=10 && string[x]!=0; tempsite.name[x]=string[x], x++);

		tempsite.name[x]=0;

		/* Site Latitude */
		fgets(string,49,fd);
		tempsite.lat=ReadBearing(string);

		/* Site Longitude */
		fgets(string,49,fd);
		tempsite.lon=ReadBearing(string);

		/* Antenna Height */
		fgets(string,49,fd);
		fclose(fd);

		/* Remove <CR> and/or <LF> from antenna height string */
		for (x=0; string[x]!=13 && string[x]!=10 && string[x]!=0; x++);

		string[x]=0;

		/* Antenna height may either be in feet or meters.
		   If the letter 'M' or 'm' is discovered in
		   the string, then this is an indication that
		   the value given is expressed in meters, and
		   must be converted to feet before exiting. */

		for (x=0; string[x]!='M' && string[x]!='m' && string[x]!=0 && x<48; x++);
		if (string[x]=='M' || string[x]=='m')
		{
			string[x]=0;
			sscanf(string,"%lf",&tempsite.alt);
			tempsite.alt*=3.28084;
		}

		else
		{
			string[x]=0;
			sscanf(string,"%lf",&tempsite.alt);
		}
	}

	return tempsite;
}

char LoadSDF_SDF(name)
char *name;
{
	/* This function reads uncompressed SPLAT Data Files (.sdf)
	   containing digital elevation model data into memory.
	   Elevation data, maximum and minimum elevations, and
	   quadrangle limits are stored in the first available
	   dem[] structure. */

	int x, y, data, indx, minlat, minlon, maxlat, maxlon;
	char found, free_slot=0, line[20], sdf_file[255], path_plus_name[255];
	FILE *fd;

	for (x=0; name[x]!='.' && name[x]!=0 && x<250; x++)
		sdf_file[x]=name[x];

	sdf_file[x]=0;

	/* Parse filename for minimum latitude and longitude values */

	sscanf(sdf_file,"%d:%d:%d:%d",&minlat,&maxlat,&minlon,&maxlon);

	sdf_file[x]='.';
	sdf_file[x+1]='s';
	sdf_file[x+2]='d';
	sdf_file[x+3]='f';
	sdf_file[x+4]=0;

	/* Is it already in memory? */

	for (indx=0, found=0; indx<MAXSLOTS && found==0; indx++)
	{
		if (minlat!=0 && minlon!=0)
		{
			if (minlat==dem[indx].min_north && minlon==dem[indx].min_west && maxlat==dem[indx].max_north && maxlon==dem[indx].max_west)
				found=1;
		}
	}

	/* Is room available to load it? */

	if (found==0)
	{	
		for (indx=0, free_slot=0; indx<MAXSLOTS && free_slot==0; indx++)
			if (dem[indx].max_north==0 && dem[indx].max_west==0)
				free_slot=1;
	}

	indx--;

	if (free_slot && found==0 && indx>=0 && indx<MAXSLOTS && minlat!=0 && minlon!=0)
	{
		strncpy(path_plus_name,sdf_path,255);
		strncat(path_plus_name,sdf_file,255);

		fd=fopen(path_plus_name,"rb");

		if (fd!=NULL)
		{
			fprintf(stdout,"Loading \"%s\" into slot %d...",path_plus_name,indx+1);
			fflush(stdout);

			fgets(line,19,fd);
			sscanf(line,"%d",&dem[indx].max_west);

			fgets(line,19,fd);
			sscanf(line,"%d",&dem[indx].min_north);

			fgets(line,19,fd);
			sscanf(line,"%d",&dem[indx].min_west);

			fgets(line,19,fd);
			sscanf(line,"%d",&dem[indx].max_north);

			for (x=0; x<1200; x++)
				for (y=0; y<1200; y++)
				{
					fgets(line,19,fd);
					sscanf(line,"%d",&data);
					dem[indx].data[x][y]=data;

					if (data>dem[indx].max_el)
						dem[indx].max_el=data;

					if (dem[indx].min_el==0)
						dem[indx].min_el=data;
					else
					{
						 if (data<dem[indx].min_el)
							dem[indx].min_el=data;
					}
				}

			fclose(fd);

			if (min_elevation==0)
				min_elevation=dem[indx].min_el;

			else
			{
				if (dem[indx].min_el<min_elevation)
					min_elevation=dem[indx].min_el;
			}

			if (dem[indx].max_el>max_elevation)
				max_elevation=dem[indx].max_el;

			if (dem[indx].max_north>max_north)
				max_north=dem[indx].max_north;

			if (dem[indx].max_west>max_west)
				max_west=dem[indx].max_west;

			if (min_north==0)
				min_north=dem[indx].min_north;
			else
			{
				if (dem[indx].min_north<min_north)
					min_north=dem[indx].min_north;
			}

			if (min_west==0)
				min_west=dem[indx].min_west;
			else
			{
				if (dem[indx].min_west<min_west)
					min_west=dem[indx].min_west;
			}

			fprintf(stdout," Done!\n");
			fflush(stdout);
			return 1;
		}
		else
			return -1;
	}
	else
		return 0;
}

char *BZfgets(bzfd,length)
BZFILE *bzfd;
unsigned length;
{
	/* This function returns at most one less than 'length' number
	   of characters from a bz2 compressed file whose file descriptor
	   is pointed to by *bzfd.  In operation, a buffer is filled with
	   uncompressed data (size = BZBUFFER), which is then parsed
	   and doled out as NULL terminated character strings every time
	   this function is invoked.  A NULL string indicates an EOF
	   or error condition. */

	static int x, y, nBuf;
	static char buffer[BZBUFFER+1], output[BZBUFFER+1], opened;
	char done=0;

	if (opened!=1 && bzerror==BZ_OK)
	{
		/* First time through.  Initialize everything! */

		x=0;
		y=0;
		nBuf=0;
		opened=1;
		output[0]=0;
	}

	do
	{
		if (x==nBuf && bzerror!=BZ_STREAM_END && bzerror==BZ_OK && opened)
		{
			/* Uncompress data into a static buffer. */

			nBuf=BZ2_bzRead(&bzerror, bzfd, buffer, BZBUFFER);
			buffer[nBuf]=0;
			x=0;
		}

		/* Build a string from buffer contents */

		output[y]=buffer[x];

		if (output[y]=='\n' || output[y]==0 || y==length-1)
		{
			output[y+1]=0;
			done=1;
			y=0;
		}

		else
			y++;
		x++;

	} while (done==0);

	if (output[0]==0)
		opened=0;

	return (output);
}

char LoadSDF_BZ(name)
char *name;
{
	/* This function reads .bz2 compressed SPLAT Data Files containing
	   digital elevation model data into memory.  Elevation data,
	   maximum and minimum elevations, and quadrangle limits are
	   stored in the first available dem[] structure. */

	int x, y, data, indx, minlat, minlon, maxlat, maxlon;
	char found, free_slot=0, sdf_file[255], path_plus_name[255];
	FILE *fd;
	BZFILE *bzfd;

	for (x=0; name[x]!='.' && name[x]!=0 && x<247; x++)
		sdf_file[x]=name[x];

	sdf_file[x]=0;

	/* Parse sdf_file name for minimum latitude and longitude values */

	sscanf(sdf_file,"%d:%d:%d:%d",&minlat,&maxlat,&minlon,&maxlon);

	sdf_file[x]='.';
	sdf_file[x+1]='s';
	sdf_file[x+2]='d';
	sdf_file[x+3]='f';
	sdf_file[x+4]='.';
	sdf_file[x+5]='b';
	sdf_file[x+6]='z';
	sdf_file[x+7]='2';
	sdf_file[x+8]=0;

	/* Is it already in memory? */

	for (indx=0, found=0; indx<MAXSLOTS && found==0; indx++)
	{
		if (minlat!=0 && minlon!=0)
		{
			if (minlat==dem[indx].min_north && minlon==dem[indx].min_west && maxlat==dem[indx].max_north && maxlon==dem[indx].max_west)
				found=1;
		}
	}

	/* Is room available to load it? */

	if (found==0)
	{	
		for (indx=0, free_slot=0; indx<MAXSLOTS && free_slot==0; indx++)
			if (dem[indx].max_north==0 && dem[indx].max_west==0)
				free_slot=1;
	}

	indx--;

	if (free_slot && found==0 && indx>=0 && indx<MAXSLOTS && minlat!=0 && minlon!=0)
	{
		strncpy(path_plus_name,sdf_path,255);
		strncat(path_plus_name,sdf_file,255);

		fd=fopen(path_plus_name,"rb");
		bzfd=BZ2_bzReadOpen(&bzerror,fd,0,0,NULL,0);

		if (fd!=NULL && bzerror==BZ_OK)
		{
			fprintf(stdout,"Loading \"%s\" into slot %d...",path_plus_name,indx+1);
			fflush(stdout);

			sscanf(BZfgets(bzfd,255),"%d",&dem[indx].max_west);
			sscanf(BZfgets(bzfd,255),"%d",&dem[indx].min_north);
			sscanf(BZfgets(bzfd,255),"%d",&dem[indx].min_west);
			sscanf(BZfgets(bzfd,255),"%d",&dem[indx].max_north);
	
			for (x=0; x<1200; x++)
				for (y=0; y<1200; y++)
				{
					sscanf(BZfgets(bzfd,20),"%d",&data);
					dem[indx].data[x][y]=data;

					if (data>dem[indx].max_el)
						dem[indx].max_el=data;

					if (dem[indx].min_el==0)
						dem[indx].min_el=data;
					else
					{
						 if (data<dem[indx].min_el)
							dem[indx].min_el=data;
					}
				}

			fclose(fd);
			BZ2_bzReadClose(&bzerror,bzfd);

			if (min_elevation==0)
				min_elevation=dem[indx].min_el;

			else
			{
				if (dem[indx].min_el<min_elevation)
					min_elevation=dem[indx].min_el;
			}
	
			if (dem[indx].max_el>max_elevation)
				max_elevation=dem[indx].max_el;

			if (dem[indx].max_north>max_north)
				max_north=dem[indx].max_north;

			if (dem[indx].max_west>max_west)
				max_west=dem[indx].max_west;

			if (min_north==0)
				min_north=dem[indx].min_north;
			else
			{
				if (dem[indx].min_north<min_north)
					min_north=dem[indx].min_north;
			}

			if (min_west==0)
				min_west=dem[indx].min_west;
			else
			{
				if (dem[indx].min_west<min_west)
					min_west=dem[indx].min_west;
			}

			fprintf(stdout," Done!\n");
			fflush(stdout);
			return 1;
		}
		else
			return -1;
	}
	else
		return 0;
}

char LoadSDF(name)
char *name;
{
	/* This function loads the requested SDF file from the filesystem.
	   It first tries to invoke the LoadSDF_SDF() function to load an
	   uncompressed SDF file (since uncompressed files load slightly
	   faster).  If that attempt fails, then it tries to load a
	   compressed SDF file by invoking the LoadSDF_BZ() function.
	   If that fails, then we can assume that no elevation data
	   exists for the region requested, and that the region
	   requested must be entirely over water. */

	int x, y, indx, minlat, minlon, maxlat, maxlon;
	char found, free_slot=0, return_value=-1;

	/* Try to load an uncompressed SDF first. */

	return_value=LoadSDF_SDF(name);

	/* If that fails, try loading a compressed SDF. */

	if (return_value==0 || return_value==-1)
		return_value=LoadSDF_BZ(name);

	/* If neither format can be found, then assume the area is water. */

	if (return_value==0 || return_value==-1)
	{
		/* Parse SDF name for minimum latitude and longitude values */

		sscanf(name,"%d:%d:%d:%d",&minlat,&maxlat,&minlon,&maxlon);

		/* Is it already in memory? */

		for (indx=0, found=0; indx<MAXSLOTS && found==0; indx++)
		{
			if (minlat!=0 && minlon!=0)
			{
				if (minlat==dem[indx].min_north && minlon==dem[indx].min_west && maxlat==dem[indx].max_north && maxlon==dem[indx].max_west)
					found=1;
			}
		}

		/* Is room available to load it? */

		if (found==0)
		{	
			for (indx=0, free_slot=0; indx<MAXSLOTS && free_slot==0; indx++)
				if (dem[indx].max_north==0 && dem[indx].max_west==0)
					free_slot=1;
		}

		indx--;

		if (free_slot && found==0 && indx>=0 && indx<MAXSLOTS && minlat!=0 && minlon!=0)
		{
			fprintf(stdout,"Filling \"%s\" as water into slot %d...",name,indx+1);
			fflush(stdout);

			dem[indx].max_west=maxlon;
			dem[indx].min_north=minlat;
			dem[indx].min_west=minlon;
			dem[indx].max_north=maxlat;

			for (x=0; x<1200; x++)
				for (y=0; y<1200; y++)
				{
					dem[indx].data[x][y]=0;

					if (dem[indx].min_el>0)
						dem[indx].min_el=0;
				}

			if (dem[indx].min_el<min_elevation)
				min_elevation=dem[indx].min_el;

			if (dem[indx].max_el>max_elevation)
				max_elevation=dem[indx].max_el;

			if (dem[indx].max_north>max_north)
				max_north=dem[indx].max_north;

			if (dem[indx].max_west>max_west)
				max_west=dem[indx].max_west;

			if (min_north==0)
				min_north=dem[indx].min_north;
			else
			{
				if (dem[indx].min_north<min_north)
					min_north=dem[indx].min_north;
			}

			if (min_west==0)
				min_west=dem[indx].min_west;
			else
			{
				if (dem[indx].min_west<min_west)
					min_west=dem[indx].min_west;
			}

			fprintf(stdout," Done!\n");
			fflush(stdout);

			return_value=1;
		}
	}

	return return_value;
}

void LoadCities(filename)
char *filename;
{
	/* This function reads SPLAT! city/site files, and plots
	   the locations and names of the cities and site locations
	   read on topographic maps generated by SPLAT! */

	int x, y, z;
	char input[80], str[3][80];
	struct site city_site;
	FILE *fd=NULL;

	fd=fopen(filename,"r");

	if (fd!=NULL)
	{
		fgets(input,78,fd);

		fprintf(stdout,"Reading \"%s\"... ",filename);
		fflush(stdout);

		while (fd!=NULL && feof(fd)==0)
		{
			/* Parse line for name, latitude, and longitude */

			for (x=0, y=0, z=0; x<78 && input[x]!=0 && z<3; x++)
			{
				if (input[x]!=',' && y<78)
				{
					str[z][y]=input[x];
					y++;
				}

				else
				{
					str[z][y]=0;
					z++;
					y=0;
				}
			}

			strncpy(city_site.name,str[0],49);
			city_site.lat=ReadBearing(str[1]);
			city_site.lon=ReadBearing(str[2]);
			city_site.alt=0.0;

			PlaceMarker(city_site);

			fgets(input,78,fd);
		}

		fclose(fd);
		fprintf(stdout,"Done!\n");
		fflush(stdout);
	}
	else
		fprintf(stderr,"*** ERROR: \"%s\": not found!\n",filename);
}

void LoadBoundaries(filename)
char *filename;
{
	/* This function reads Cartographic Boundary Files available from
	   the U.S. Census Bureau, and plots the data contained in those
	   files on the PPM Map generated by SPLAT!.  Such files contain
	   the coordinates that describe the boundaries of cities, counties,
	   and states. */

	int x;
	double lat0, lon0, lat1, lon1;
	char string[80];
	struct site source, destination;
	FILE *fd=NULL;

	fd=fopen(filename,"r");

	if (fd!=NULL)
	{
		fgets(string,78,fd);

		fprintf(stdout,"Reading \"%s\"... ",filename);
		fflush(stdout);

		do
		{
			fgets(string,78,fd);
			sscanf(string,"%lf %lf", &lon0, &lat0);
			fgets(string,78,fd);

			do
			{
				sscanf(string,"%lf %lf", &lon1, &lat1);
			
				lon0=fabs(lon0);
				lon1=fabs(lon1);

				source.lat=lat0;
				source.lon=lon0;
				destination.lat=lat1;
				destination.lon=lon1;

				ReadPath(source,destination);

				for (x=0; x<path.length; x++)
					OrMask(path.lat[x],path.lon[x],4);

				lat0=lat1;
				lon0=lon1;

				fgets(string,78,fd);

			} while (strncmp(string,"END",3)!=0 && feof(fd)==0);

			fgets(string,78,fd);

		} while (strncmp(string,"END",3)!=0 && feof(fd)==0);

		fclose(fd);

		fprintf(stdout,"Done!\n");
		fflush(stdout);
	}
	else
		fprintf(stderr,"*** ERROR: \"%s\": not found!\n",filename);
}

struct site los(source,destination)
struct site source, destination;
{
	/* This function determines whether a line-of-sight path
	   unobstructed by terrain exists between source (transmitter)
	   and destination (receiver) based on the geographical
	   locations of the two sites, their respective antenna
	   heights above ground, and the terrain between them.
	   A site structure is returned upon completion.  If the
	   first character of site.name is ' ', then a clear path
	   exists between source and destination.  If the first
	   character is '*', then an obstruction exists, and the
	   site.lat and site.lon elements of the structure provide
	   the geographical location of the obstruction. */
	   
	int x;
	struct site test, blockage;
	register double distance, tx_alt, rx_alt,
		 cos_xmtr_angle, cos_test_angle, test_alt;
	char block;

	ReadPath(source,destination);

	distance=5280.0*Distance(source,destination);
	tx_alt=EARTHRADIUS+source.alt+GetElevation(source);
	rx_alt=EARTHRADIUS+destination.alt+GetElevation(destination);

	/* Elevation angle of the xmtr (source) as seen by the rcvr */

	cos_xmtr_angle=((rx_alt*rx_alt)+(distance*distance)-(tx_alt*tx_alt))/(2.0*rx_alt*distance);

	/* Determine the elevation angle of each discrete location
	   along the path between the receiver and transmitter.

	   Since obstructions are more likely due to terrain effects
	   closest to the receiver rather than farther away, we start
	   looking for potential obstructions from the receiver's
	   location, and work our way towards the transmitter.
	   This loop is broken when the first obstruction is
	   detected.  If we can travel all the way to the transmitter
	   without detecting an obstruction, then we have a clear
	   unobstructed path between transmitter and receiver. */

	for (x=path.length-1, block=0; x>0 && block==0; x--)
	{
		/* Build a structure for each test
		   point along the path to be surveyed. */

		test.lat=path.lat[x];
		test.lon=path.lon[x];

		/* Measure the distance between the
		   test point and the receiver locations */

		distance=5280.0*Distance(test,destination);
		test_alt=EARTHRADIUS+path.elevation[x];

		/* Determine the cosine of the elevation of the test
		   point as seen from the location of the receiver */

		cos_test_angle=((rx_alt*rx_alt)+(distance*distance)-(test_alt*test_alt))/(2.0*rx_alt*distance);

		/* If the elevation angle to the test point (as seen from
		   the receiver) is greater than the elevation angle to the
		   transmitter (as seen by the receiver), then we have a
		   path obstructed by terrain.  Note: Since we're comparing
	   	   the cosines of these angles rather than the angles
		   themselves (eliminating the call to acos() saves
		   considerable time), the following "if" statement is
		   reversed from what it would normally be if the angles
		   were compared. */

		if (cos_xmtr_angle>cos_test_angle)
		{
			block=1;
			blockage.lat=path.lat[x];
			blockage.lon=path.lon[x];
			blockage.alt=path.elevation[x];
			blockage.name[0]='*';
		}
	}

	if (block==0)
	{
		blockage.lat=0.0;
		blockage.lon=0.0;
		blockage.alt=0.0;
		blockage.name[0]=' ';
	}

	return blockage;
}

void PlotPath(source,destination,mask_value)
struct site source, destination;
char mask_value;
{
	/* This function analyzes the path between the source and
	   destination locations.  It determines which points along
	   the path have line-of-sight visibility to the source.
	   Points along with path having line-of-sight visibility
	   to the source at an AGL altitude equal to that of the
	   destination location are stored by setting bit 1 in the
	   mask[][] array, which are displayed in green when PPM
	   maps are later generated by SPLAT!. */

	char block;
	int x, y;
	register double cos_xmtr_angle, cos_test_angle, test_alt;
	double distance, rx_alt, tx_alt;

	ReadPath(source,destination);

	for (y=0; y<path.length; y++)
	{
		/* Test this point only if it hasn't been already
		   tested and found to be free of obstructions. */

		if ((GetMask(path.lat[y],path.lon[y])&mask_value)==0)
		{
			distance=5280.0*path.distance[y];
			tx_alt=EARTHRADIUS+source.alt+path.elevation[0];
			rx_alt=EARTHRADIUS+destination.alt+path.elevation[y];

			/* Calculate the cosine of the elevation of the
			   transmitter as seen at the temp rx point. */

			cos_xmtr_angle=((rx_alt*rx_alt)+(distance*distance)-(tx_alt*tx_alt))/(2.0*rx_alt*distance);

			for (x=y, block=0; x>=0 && block==0; x--)
			{
				distance=5280.0*(path.distance[y]-path.distance[x]);
				test_alt=EARTHRADIUS+path.elevation[x];

				cos_test_angle=((rx_alt*rx_alt)+(distance*distance)-(test_alt*test_alt))/(2.0*rx_alt*distance);

				/* Compare these two angles to determine if
				   a blockage exists.  Since we're comparing
				   the cosines of these angles rather than
				   the angles themselves, the following "if"
				   statement is reversed from what it would
				   be if the actual angles were compared. */

				if (cos_xmtr_angle>cos_test_angle)
					block=1;
			}

			if (block==0)
				OrMask(path.lat[y],path.lon[y],mask_value);
		}
	}
}

void PlotCoverage(source, altitude)
struct site source;
double altitude;
{
	/* This function performs a 360 degree sweep around the
	   transmitter site (source location), and plots the
	   line-of-sight coverage of the transmitter on the SPLAT!
	   generated topographic map based on a receiver located
	   at the specified altitude (in feet AGL).  Results
	   are stored in memory, and written out in the form
	   of a topographic map when the WritePPM() function
	   is later invoked. */

	double lat, lon, one_pixel;
	static unsigned char mask_value;
	int z, count;
	struct site edge;

	/* Initialize mask_value */

	if (mask_value!=8 && mask_value!=16 && mask_value!=32)
		mask_value=1;

	one_pixel=1.0/1200.0;

	count=0;	
	fprintf(stdout,"\nPlotting coverage of %s with a receive antenna at %.0f feet AGL:\n 0%c to  25%c",source.name,altitude,37,37);
	fflush(stdout);

	/* 18.461538=1200 pixels/degree divided by 65 loops per '.' printed. */

	z=(int)(18.461538*(max_west-min_west));

	for (lon=min_west; lon<=max_west; lon+=one_pixel)
	{
		edge.lat=max_north;
		edge.lon=lon;
		edge.alt=altitude;

		PlotPath(source,edge,mask_value);
		count++;

		if (count==z) 
		{
			fprintf(stdout,".");
			fflush(stdout);
			count=0;
		}
	}

	count=0;
	fprintf(stdout,"\n25%c to  50%c",37,37);
	fflush(stdout);
	
	z=(int)(18.461538*(max_north-min_north));

	for (lat=max_north; lat>=min_north; lat-=one_pixel)
	{
		edge.lat=lat;
		edge.lon=min_west;
		edge.alt=altitude;

		PlotPath(source,edge,mask_value);
		count++;

		if (count==z) 
		{
			fprintf(stdout,".");
			fflush(stdout);
			count=0;
		}
	}

	count=0;
	fprintf(stdout,"\n50%c to  75%c",37,37);
	fflush(stdout);

	z=(int)(18.461538*(max_west-min_west));

	for (lon=min_west; lon<=max_west; lon+=one_pixel)
	{
		edge.lat=min_north;
		edge.lon=lon;
		edge.alt=altitude;

		PlotPath(source,edge,mask_value);
		count++;

		if (count==z)
		{
			fprintf(stdout,".");
			fflush(stdout);
			count=0;
		}
	}

	count=0;
	fprintf(stdout,"\n75%c to 100%c",37,37);
	fflush(stdout);
	
	z=(int)(18.461538*(max_north-min_north));

	for (lat=min_north; lat<=max_north; lat+=one_pixel)
	{
		edge.lat=lat;
		edge.lon=max_west;
		edge.alt=altitude;

		PlotPath(source,edge,mask_value);
		count++;

		if (count==z)
		{
			fprintf(stdout,".");
			fflush(stdout);
			count=0;
		}
	}

	fprintf(stdout,"\nDone!\n");
	fflush(stdout);

	/* Assign next mask value */

	switch (mask_value)
	{
		case 1:
			mask_value=8;
			break;

		case 8:
			mask_value=16;
			break;

		case 16:
			mask_value=32;
	}
}

void WritePPM(filename)
char *filename;
{
	/* This function generates a topographic map in Portable Pix Map
	   (PPM) format based on logarithmically scaled topology data,
	   as well as the content of flags held in the mask[][] array.
	   The image created is rotated counter-clockwise 90 degrees
	   from its representation in dem[][] so that north points
	   up and east points right in the image generated. */

	int indx, x, x0, y0, minlat, minlon;
	unsigned width, height, output;
	unsigned char found, mask;
	char mapfile[255];
	double conversion, lat, lon, one_over_gamma, one_pixel;
	FILE *fd;

	one_pixel=1.0/1200.0;
	one_over_gamma=1.0/GAMMA;
	conversion=255.0/pow((double)(max_elevation-min_elevation),one_over_gamma);

	width=1200*(max_west-min_west);
	height=1200*(max_north-min_north);

	if (filename[0]==0)
		strncpy(mapfile, "map.ppm\0",8);
	else
	{
		for (x=0; filename[x]!='.' && filename[x]!=0 && x<250; x++)
			mapfile[x]=filename[x];

		mapfile[x]='.';
		mapfile[x+1]='p';
		mapfile[x+2]='p';
		mapfile[x+3]='m';
		mapfile[x+4]=0;
	}

	fd=fopen(mapfile,"wb");

	fprintf(fd,"P6\n%u %u\n255\n",width,height);

	fprintf(stdout,"\nWriting \"%s\" (%ux%u pixmap image)... ",mapfile,width,height);
	fflush(stdout);

	for (lat=(double)max_north; lat>=(double)min_north; lat-=one_pixel)
	{
		for (lon=(double)max_west; lon>=(double)min_west; lon-=one_pixel)
		{
			minlat=(int)floor(lat);
			minlon=(int)floor(lon);

			for (indx=0, found=0; indx<MAXSLOTS && found==0;)
				if (minlat==dem[indx].min_north && minlon==dem[indx].min_west)
					found=1;
				else
					indx++;
			if (found)
			{
				x0=(int)((1199.0*(lat-floor(lat)))+0.5);
				y0=(int)((1199.0*(lon-floor(lon)))+0.5);

				mask=dem[indx].mask[x0][y0];

				if (mask&2)
					/* Text Labels: Red */
					fprintf(fd,"%c%c%c",255,0,0);

				else if (mask&4)
					/* County Boundaries: Light Cyan */
					fprintf(fd,"%c%c%c",128,128,255);

				else switch (mask&57)
				{
					case 1:
					/* TX1: Green */
					fprintf(fd,"%c%c%c",0,255,0);
					break;

					case 8:
					/* TX2: Cyan */
					fprintf(fd,"%c%c%c",0,255,255);
					break;

					case 9:
					/* TX1 + TX2: Yellow */
					fprintf(fd,"%c%c%c",255,255,0);
					break;

					case 16:
					/* TX3: Medium Violet */
					fprintf(fd,"%c%c%c",147,112,219);
					break;

					case 17:
					/* TX1 + TX3: Pink */
					fprintf(fd,"%c%c%c",255,192,203);
					break;

					case 24:
					/* TX2 + TX3: Orange */
					fprintf(fd,"%c%c%c",255,165,0);
					break;

				 	case 25:
					/* TX1 + TX2 + TX3: Dark Green */
					fprintf(fd,"%c%c%c",0,100,0);
					break;

					case 32:
					/* TX4: Sienna 1 */
					fprintf(fd,"%c%c%c",255,130,71);
					break;

					case 33:
					/* TX1 + TX4: Green Yellow */
					fprintf(fd,"%c%c%c",173,255,47);
					break;

					case 40:
					/* TX2 + TX4: Dark Sea Green 1 */
					fprintf(fd,"%c%c%c",193,255,193);
					break;

					case 41:
					/* TX1 + TX2 + TX4: Blanched Almond */
					fprintf(fd,"%c%c%c",255,235,205);
					break;

					case 48:
					/* TX3 + TX4: Dark Turquoise */
					fprintf(fd,"%c%c%c",0,206,209);
					break;

					case 49:
					/* TX1 + TX3 + TX4: Medium Spring Green */
					fprintf(fd,"%c%c%c",0,250,154);
					break;

					case 56:
					/* TX2 + TX3 + TX4: Tan */
					fprintf(fd,"%c%c%c",210,180,140);
					break;

					case 57:
					/* TX1 + TX2 + TX3 + TX4: Gold2 */
					fprintf(fd,"%c%c%c",238,201,0);
					break;

					default:
					/* Water: Medium Blue */
					if (dem[indx].data[x0][y0]==0)
						fprintf(fd,"%c%c%c",0,0,170);
					else
					{
						/* Elevation: Greyscale */
						output=(unsigned)0.5+pow((double)(dem[indx].data[x0][y0]-min_elevation),one_over_gamma)*conversion;
						fprintf(fd,"%c%c%c",output,output,output);
					}
				}
			}
		}
	}

	fclose(fd);
	fprintf(stdout,"Done!\n");
	fflush(stdout);
}

void GraphTerrain(source,destination,name)
struct site source, destination;
char *name;
{
	/* This function invokes gnuplot to generate an appropriate
	   output file indicating the terrain profile between the source
	   and destination locations.  "filename" is the name assigned
	   to the output file generated by gnuplot.  The filename extension
	   is used to set gnuplot's terminal setting and output file type.
	   If no extension is found, .gif is assumed.  */

	int x, y, z;
	char filename[255], term[15], ext[15];
	FILE *fd=NULL;

	ReadPath(destination,source);

	fd=fopen("profile.gp","wb");

	for (x=0; x<path.length; x++)
		fprintf(fd,"%f\t%f\n",path.distance[x],path.elevation[x]);

	fclose(fd);

	if (name[0]==0)
	{
		/* Default filename and output file type */

		strncpy(filename,"profile\0",8);
		strncpy(term,"gif\0",4);
		strncpy(ext,"gif\0",4);
	}

	else
	{
		/* Grab extension and terminal type from "name" */

		for (x=0; name[x]!='.' && name[x]!=0 && x<254; x++)
			filename[x]=name[x];

		if (name[x]=='.')
		{
			for (y=0, z=x, x++; name[x]!=0 && x<254 && y<14; x++, y++)
			{
				term[y]=tolower(name[x]);
				ext[y]=term[y];
			}

			ext[y]=0;
			term[y]=0;
			filename[z]=0;
		}

		else
		{	/* No extension -- Default is gif */

			filename[x]=0;
			strncpy(term,"gif\0",4);
			strncpy(ext,"gif\0",4);
		}
	}

	/* Either .ps or .postscript may be used
	   as an extension for postscript output. */

	if (strncmp(term,"postscript",10)==0)
		strncpy(ext,"ps\0",3);

	else if (strncmp(ext,"ps",2)==0)
		strncpy(term,"postscript\0",11);

	fprintf(stdout,"Writing \"%s.%s\"...",filename,ext);
	fflush(stdout);

	fd=fopen("splat.gp","w");
	fprintf(fd,"set grid\n");
	fprintf(fd,"set autoscale\n");
	fprintf(fd,"set term %s\n",term);
	fprintf(fd,"set title \"SPLAT! Terrain Profile\"\n");
	fprintf(fd,"set xlabel \"Distance Between %s and %s (miles)\"\n",destination.name,source.name);
	fprintf(fd,"set ylabel \"Ground Elevation Above Sea Level (feet)\"\n");
	fprintf(fd,"set output \"%s.%s\"\n",filename,ext);
	fprintf(fd,"plot \"profile.gp\" title \"\" with lines\n");
	fclose(fd);
			
	x=system("gnuplot splat.gp");

	if (x!=-1)
	{
		unlink("splat.gp");
		unlink("profile.gp");
		fprintf(stdout," Done!\n");
		fflush(stdout);
	}

	else
		fprintf(stderr,"\n*** ERROR: Error occurred invoking gnuplot!\n");
}

void GraphElevation(source,destination,name)
struct site source, destination;
char *name;
{
	/* This function invokes gnuplot to generate an appropriate
	   output file indicating the terrain profile between the source
	   and destination locations.  "filename" is the name assigned
	   to the output file generated by gnuplot.  The filename extension
	   is used to set gnuplot's terminal setting and output file type.
	   If no extension is found, .gif is assumed.  */

	int x, y, z;
	char filename[255], term[15], ext[15];
	double angle, refangle, maxangle=-90.0;
	struct site remote;
	FILE *fd=NULL, *fd2=NULL;

	ReadPath(destination,source);  /* destination=RX, source=TX */
	refangle=ElevationAngle(destination,source);

	fd=fopen("profile.gp","wb");
	fd2=fopen("reference.gp","wb");

	for (x=1; x<path.length-1; x++)
	{
		remote.lat=path.lat[x];
		remote.lon=path.lon[x];
		remote.alt=0.0;
		angle=ElevationAngle(destination,remote);
		fprintf(fd,"%f\t%f\n",path.distance[x],angle);
		fprintf(fd2,"%f\t%f\n",path.distance[x],refangle);

		if (angle>maxangle)
			maxangle=angle;
	}

	fprintf(fd,"%f\t%f\n",path.distance[path.length-1],refangle);
	fprintf(fd2,"%f\t%f\n",path.distance[path.length-1],refangle);

	fclose(fd);
	fclose(fd2);

	if (name[0]==0)
	{
		/* Default filename and output file type */

		strncpy(filename,"profile\0",8);
		strncpy(term,"gif\0",4);
		strncpy(ext,"gif\0",4);
	}

	else
	{
		/* Grab extension and terminal type from "name" */

		for (x=0; name[x]!='.' && name[x]!=0 && x<254; x++)
			filename[x]=name[x];

		if (name[x]=='.')
		{
			for (y=0, z=x, x++; name[x]!=0 && x<254 && y<14; x++, y++)
			{
				term[y]=tolower(name[x]);
				ext[y]=term[y];
			}

			ext[y]=0;
			term[y]=0;
			filename[z]=0;
		}

		else
		{	/* No extension -- Default is gif */

			filename[x]=0;
			strncpy(term,"gif\0",4);
			strncpy(ext,"gif\0",4);
		}
	}

	/* Either .ps or .postscript may be used
	   as an extension for postscript output. */

	if (strncmp(term,"postscript",10)==0)
		strncpy(ext,"ps\0",3);

	else if (strncmp(ext,"ps",2)==0)
		strncpy(term,"postscript\0",11);

	fprintf(stdout,"Writing \"%s.%s\"...",filename,ext);
	fflush(stdout);

	fd=fopen("splat.gp","w");

	fprintf(fd,"set grid\n");
	fprintf(fd,"set yrange [%2.3f to %2.3f]\n", (-fabs(refangle)-0.25), maxangle+0.25);
	fprintf(fd,"set term %s\n",term);
	fprintf(fd,"set title \"SPLAT! Elevation Profile\"\n");
	fprintf(fd,"set xlabel \"Distance Between %s and %s (miles)\"\n",destination.name,source.name);
	fprintf(fd,"set ylabel \"Elevation Angle Along Path Between %s and %s (degrees)\"\n",destination.name,source.name);
	fprintf(fd,"set output \"%s.%s\"\n",filename,ext);
	fprintf(fd,"plot \"profile.gp\" title \"Real Earth Profile\" with lines, \"reference.gp\" title \"Line Of Sight Path\" with lines\n");

	fclose(fd);
			
	x=system("gnuplot splat.gp");

	if (x!=-1)
	{
		unlink("splat.gp");
		unlink("profile.gp");
		unlink("reference.gp");
		fprintf(stdout," Done!\n");
		fflush(stdout);
	}

	else
		fprintf(stderr,"\n*** ERROR: Error occurred invoking gnuplot!\n");
}

void ObstructionReport(xmtr, rcvr, report)
struct site xmtr, rcvr;
char report;
{
	struct site result, result2, new_site;
	double angle, haavt;
	unsigned char block, report_name[80], string[80];
	int x;
	FILE *fd;

	sprintf(report_name,"%s-to-%s.txt",xmtr.name,rcvr.name);

	for (x=0; report_name[x]!=0; x++)
		if (report_name[x]==32 || report_name[x]==17 || report_name[x]==92 || report_name[x]==42 || report_name[x]==47)
			report_name[x]='_';	

	fd=fopen(report_name,"w");

	fprintf(fd,"\n\t\t--==[ SPLAT! v%s Obstruction Report ]==--\n\n",splat_version);
	fprintf(fd,"Analysis of line-of-sight path conditions between %s and %s:\n",xmtr.name, rcvr.name);
	fprintf(fd,"\n-------------------------------------------------------------------------\n\n");
	fprintf(fd,"Transmitter site: %s\n",xmtr.name);
	fprintf(fd,"Site location: %.4f North / %.4f West",xmtr.lat, xmtr.lon);
	fprintf(fd, " (%s N / ", dec2dms(xmtr.lat));
	fprintf(fd, "%s W)\n", dec2dms(xmtr.lon));
	fprintf(fd,"Ground elevation: %.2f feet AMSL\n",GetElevation(xmtr));
	fprintf(fd,"Antenna height: %.2f feet AGL / %.2f feet AMSL\n",xmtr.alt, xmtr.alt+GetElevation(xmtr));

	haavt=haat(xmtr);

	if (haavt>-4999.0)
		fprintf(fd,"Antenna height above average terrain: %.2f feet\n",haavt);

	fprintf(fd,"Distance to %s: %.2f miles.\n",rcvr.name,Distance(xmtr,rcvr));
	fprintf(fd,"Azimuth to %s: %.2f degrees.\n",rcvr.name,Azimuth(xmtr,rcvr));

	angle=ElevationAngle(xmtr,rcvr);

	if (angle>=0.0)
		fprintf(fd,"Angle of elevation between %s and %s: %+.4f degrees.\n",xmtr.name,rcvr.name,angle);

	if (angle<0.0)
		fprintf(fd,"Angle of depression between %s and %s: %+.4f degrees.\n",xmtr.name,rcvr.name,angle);

	fprintf(fd,"\n-------------------------------------------------------------------------\n\n");

	/* Receiver */

	fprintf(fd,"Receiver site: %s\n",rcvr.name);
	fprintf(fd,"Site location: %.4f North / %.4f West",rcvr.lat, rcvr.lon);
	fprintf(fd, " (%s N / ", dec2dms(rcvr.lat));
	fprintf(fd, "%s W)\n", dec2dms(rcvr.lon));
	fprintf(fd,"Ground elevation: %.2f feet AMSL\n",GetElevation(rcvr));
	fprintf(fd,"Antenna height: %.2f feet AGL / %.2f feet AMSL\n",rcvr.alt, rcvr.alt+GetElevation(rcvr));

	haavt=haat(rcvr);

	if (haavt>-4999.0)
		fprintf(fd,"Antenna height above average terrain: %.2f feet\n",haavt);

	fprintf(fd,"Distance to %s: %.2f miles.\n",xmtr.name,Distance(xmtr,rcvr));
	fprintf(fd,"Azimuth to %s: %.2f degrees.\n",xmtr.name,Azimuth(rcvr,xmtr));

	angle=ElevationAngle(rcvr,xmtr);

	if (angle>=0.0)
		fprintf(fd,"Angle of elevation between %s and %s: %+.4f degrees.\n",rcvr.name,xmtr.name,angle);

	if (angle<0.0)
		fprintf(fd,"Angle of depression between %s and %s: %+.4f degrees.\n",rcvr.name,xmtr.name,angle);

	fprintf(fd,"\n-------------------------------------------------------------------------\n\n");

	if (report=='y')
	{
		/* Write an Obstruction Report */

		new_site=rcvr;
		result=los(xmtr,rcvr);
		block=result.name[0];

		if (block=='*')
			fprintf(fd,"SPLAT! detected obstructions at:\n\n");

		while (block=='*')
		{
			if (result.lat!=result2.lat || result.lon!=result2.lon || result.alt!=result2.alt)
				fprintf(fd,"\t%.4f N, %.4f W, %5.2f miles, %6.2f feet AMSL.\n",result.lat, result.lon, Distance(rcvr,result), result.alt);

			result2=result;
			new_site.alt+=1.0;

			/* Can you hear me now? :-) */

			result=los(xmtr,new_site);
			block=result.name[0];
		}

		if (new_site.alt!=rcvr.alt)
			fprintf(fd,"\nAntenna at %s must be raised to at least %.2f feet AGL\nto clear all obstructions detected by SPLAT!\n\n",rcvr.name, new_site.alt);
		else
			fprintf(fd,"No obstructions due to terrain were detected by SPLAT!\n\n");
	}

	fclose(fd);

	/* Display report to terminal */

	fd=fopen(report_name,"r");

	fgets(string,80,fd);

	while (feof(fd)==0)
	{
		fprintf(stdout,"%s",string);
		fgets(string,80,fd);
	}

	fclose(fd);
	fprintf(stdout,"Obstruction report written to: \"%s\"\n",report_name);
	fflush(stdout);
}

void SiteReport(xmtr)
struct site xmtr;
{
	char report_name[80];
	double terrain;
	int x, azi;
	FILE *fd;

	sprintf(report_name,"%s-site_report.txt",xmtr.name);

	for (x=0; report_name[x]!=0; x++)
		if (report_name[x]==32 || report_name[x]==17 || report_name[x]==92 || report_name[x]==42 || report_name[x]==47)
			report_name[x]='_';	

	fd=fopen(report_name,"w");

	fprintf(fd,"\n\t--==[ SPLAT! v%s Site Analysis Report For: %s ]==--\n\n",splat_version,xmtr.name);

	fprintf(fd,"---------------------------------------------------------------------------\n\n");
	fprintf(fd,"Site location: %.4f North / %.4f West",xmtr.lat, xmtr.lon);
	fprintf(fd, " (%s N / ",dec2dms(xmtr.lat));
	fprintf(fd, "%s W)\n",dec2dms(xmtr.lon));
	fprintf(fd,"Ground elevation: %.2f feet AMSL\n",GetElevation(xmtr));
	fprintf(fd,"Antenna height: %.2f feet AGL / %.2f feet AMSL\n",xmtr.alt, xmtr.alt+GetElevation(xmtr));

	terrain=haat(xmtr);

	if (terrain>-4999.0)
	{
		fprintf(fd,"Antenna height above average terrain: %.2f feet\n\n",terrain);

		/* Display the average terrain between 2 and 10 miles
		   from the transmitter site at azimuths of 0, 45, 90,
		   135, 180, 225, 270, and 315 degrees. */

		for (azi=0; azi<=315; azi+=45)
		{
			fprintf(fd,"Average terrain at %3d degrees azimuth: ",azi);
			terrain=AverageTerrain(xmtr,(double)azi,2.0,10.0);

			if (terrain>-4999.0)
				fprintf(fd,"%.2f feet AMSL\n",terrain);
			else
				fprintf(fd,"No terrain\n");
		}
	}

	fprintf(fd,"\n---------------------------------------------------------------------------\n\n");
	fclose(fd);
	fprintf(stdout,"\nSite analysis report written to: \"%s\"\n",report_name);
}

int main(argc,argv)
char argc, *argv[];
{

	int		x, y, z=0;
	unsigned char	rxlat, rxlon, txlat, txlon, min_lat,
			min_lon, max_lat, max_lon, mapfile[255],
			elevation_file[255], terrain_file[255],
			coverage=0, string[255], rxfile[255],
			txfile[255], map=0, boundary_file[5][255],
			ext[20], city_file[5][255], terrain_plot=0,
			elevation_plot=0, cities=0, bfs=0, txsites=0,
			count, west_min, west_max, north_min, north_max,
			header[80], report='y';
	double		altitude=0.0, tx_range, rx_range, deg_range, deg_limit;
	struct		site tx_site[4], rx_site;
	FILE		*fd;

	sprintf(header,"\n  --==[ SPLAT! v%s Terrain Analysis Software by KD2BD (c) 1997-2002 ]==--\n\n", splat_version);

	if (argc==1)
	{
		fprintf(stdout, "%sAvailable Options...\n\n\t-t txsite(s).qth (max of 4)\n\t-r rxsite.qth\n",header);
		fprintf(stdout,"\t-c plot coverage area(s) of TX(s) based on an RX antenna at X feet AGL\n");
		fprintf(stdout,"\t-s filename(s) of city/site file(s) to import (max of 5)\n");
		fprintf(stdout,"\t-b filename(s) of cartographic boundary file(s) to import (max of 5)\n");
		fprintf(stdout,"\t-p filename of terrain profile graph to plot\n");
		fprintf(stdout,"\t-e filename of terrain elevation graph to plot\n");
		fprintf(stdout,"\t-o filename of topographic map to generate (.ppm)\n");
		fprintf(stdout,"\t-d sdf file directory path (overrides path in ~/.splat_path file)\n");
		fprintf(stdout,"\t-n no analysis, brief report.\n\t-N no analysis, no report.\n\n");
		fprintf(stdout,"Type 'man splat', or see the documentation for more details.\n\n");
		fflush(stdout);
		exit(1);
	}

	y=argc-1;

	rxfile[0]=0;
	txfile[0]=0;
	string[0]=0;
	mapfile[0]=0;
	elevation_file[0]=0;
	terrain_file[0]=0;
	sdf_path[0]=0;
	path.length=0;
	rx_site.lat=0.0;
	rx_site.lon=0.0;

	for (x=0; x<MAXSLOTS; x++)
	{
		dem[x].min_el=0;
		dem[x].max_el=0;
		dem[x].min_north=0;
		dem[x].max_north=0;
		dem[x].min_west=0;
		dem[x].max_west=0;
	}

	/* Scan for command line arguments */

	for (x=1; x<=y; x++)
	{
		if (strcmp(argv[x],"-o")==0)
		{
			z=x+1;

			if (z<=y && argv[z][0] && argv[z][0]!='-')
				strncpy(mapfile,argv[z],253);
			map=1;
		}

		if (strcmp(argv[x],"-c")==0)
		{
			z=x+1;

			if (z<=y && argv[z][0] && argv[z][0]!='-')
			{
				sscanf(argv[z],"%lf",&altitude);
				coverage=1;
			}
		}

		if (strcmp(argv[x],"-p")==0)
		{
			z=x+1;

			if (z<=y && argv[z][0] && argv[z][0]!='-')
			{
				strncpy(terrain_file,argv[z],253);
				terrain_plot=1;
			}
		}

		if (strcmp(argv[x],"-e")==0)
		{
			z=x+1;

			if (z<=y && argv[z][0] && argv[z][0]!='-')
			{
				strncpy(elevation_file,argv[z],253);
				elevation_plot=1;
			}
		}

		if (strcmp(argv[x],"-n")==0)
		{
			if (z<=y && argv[z][0] && argv[z][0]!='-')
			{
				report='n';
				map=1;
			}
		}

		if (strcmp(argv[x],"-N")==0)
		{
			if (z<=y && argv[z][0] && argv[z][0]!='-');
			{
				report='N';
				map=1;
			}
		}

		if (strcmp(argv[x],"-d")==0)
		{
			z=x+1;

			if (z<=y && argv[z][0] && argv[z][0]!='-')
				strncpy(sdf_path,argv[z],253);
		}

		if (strcmp(argv[x],"-t")==0)
		{
			/* Read Transmitter Location */

			z=x+1;

			while (z<=y && argv[z][0] && argv[z][0]!='-' && txsites<4)
			{
				strncpy(txfile,argv[z],253);
				tx_site[txsites]=LoadQTH(txfile);
				txsites++;
				z++;
			}
			z--;
		}

		if (strcmp(argv[x],"-r")==0)
		{
			/* Read Receiver Location */

			z=x+1;

			if (z<=y && argv[z][0] && argv[z][0]!='-')
			{
				strncpy(rxfile,argv[z],253);
				rx_site=LoadQTH(rxfile);
			}
		}

		if (strcmp(argv[x],"-s")==0)
		{
			/* Read city file(s) */

			z=x+1;

			while (z<=y && argv[z][0] && argv[z][0]!='-' && cities<5)
			{
				strncpy(city_file[cities],argv[z],253);
				cities++;
				z++;
			}
			z--;
		}

		if (strcmp(argv[x],"-b")==0)
		{
			/* Read Boundary File(s) */

			z=x+1;

			while (z<=y && argv[z][0] && argv[z][0]!='-' && bfs<5)
			{
				strncpy(boundary_file[bfs],argv[z],253);
				bfs++;
				z++;
			}
			z--;
		}
	}

	/* Perform some error checking on the arguments
	   and switches parsed from the command-line.
	   If an error is encountered, print a message
	   and exit gracefully. */

	if (txsites==0)
	{
		fprintf(stderr,"\n%c*** ERROR: No transmitter site(s) specified!\n\n",7);
		exit (-1);
	}

	for (x=0, y=0; x<txsites; x++)
	{
		if (tx_site[x].lat==0.0 && tx_site[x].lon==0.0)
		{
			fprintf(stderr,"\n*** ERROR: Transmitter site #%d not found!",x+1);
			y++;
		}
	}

	if (y)
	{
		fprintf(stderr,"%c\n\n",7);
		exit (-1);
	}

	if (coverage==0 && rx_site.lat==0.0 && rx_site.lon==0.0)
	{
		fprintf(stderr,"\n%c*** ERROR: No receiver site found or specified!\n\n",7);
		exit (-1);
	}

	/* No errors were detected.  Whew!  :-) */

	/* If no SDF path was specified on the command line (-d), check
	   for a path specified in the $HOME/.splat_path file.  If the
	   file is not found, then sdf_path[] remains NULL, and the
	   current working directory is assumed to contain the SDF
	   files. */

	if (sdf_path[0]==0)
	{
		sprintf(string,"%s/.splat_path",getenv("HOME"));
		fd=fopen(string,"r");

		if (fd!=NULL)
		{
			fgets(string,253,fd);

			/* Remove <CR> and/or <LF> from string */
			for (x=0; string[x]!=13 && string[x]!=10 && string[x]!=0 && x<253; x++);
			string[x]=0;

			strncpy(sdf_path,string,253);

			fclose(fd);
		}
	}

	/* Ensure a trailing '/' is present in sdf_path */

	if (sdf_path[0])
	{
		x=strlen(sdf_path);

		if (sdf_path[x-1]!='/' && x!=0)
		{
			sdf_path[x]='/';
			sdf_path[x+1]=0;
		}
	}

	fprintf(stdout,"%s",header);
	fflush(stdout);

	x=0;
	y=0;

	min_lat=0;
	max_lat=0;
	min_lon=0;
	max_lon=0;

	rxlat=(unsigned char)floor(rx_site.lat);
	rxlon=(unsigned char)floor(rx_site.lon);

	if (rxlat!=0)
	{
		if (min_lat==0)
			min_lat=rxlat;

		else if (rxlat<min_lat)
			min_lat=rxlat;
	}

	if (rxlon!=0)
	{
		if (min_lon==0)
			min_lon=rxlon;

		else if (rxlon<min_lon)
			min_lon=rxlon;
	}

	if (rxlat>max_lat)
		max_lat=rxlat;

	if (rxlon>max_lon)
		max_lon=rxlon;

	for (y=0, z=0; z<txsites; z++)
	{
		txlat=(unsigned char)floor(tx_site[z].lat);
		txlon=(unsigned char)floor(tx_site[z].lon);

		if (txlat!=0)
		{
			if (min_lat==0)
				min_lat=txlat;

			else if (txlat<min_lat)
				min_lat=txlat;
		}

		if (txlon!=0)
		{
			if (min_lon==0)
				min_lon=txlon;

			else if (txlon<min_lon)
				min_lon=txlon;
		}

		if (txlat>max_lat)
			max_lat=txlat;

		if (txlon>max_lon)
			max_lon=txlon;
	}

	if (min_lat!=0 && min_lon!=0 && max_lat!=0 && max_lon!=0)
	{
		for (y=min_lon; y<=max_lon; y++)
			for (x=min_lat; x<=max_lat; x++)
			{
				sprintf(string,"%u:%u:%u:%u",x, x+1, y, y+1);
				LoadSDF(string);
			}
	}

	if (coverage)
	{
		for (z=0; z<txsites; z++)
		{
			/* "Ball park" estimates used to load any additional
			   SDF files required to conduct this analysis. */

			tx_range=sqrt(1.5*(tx_site[z].alt+GetElevation(tx_site[z])));
			rx_range=sqrt(1.5*altitude);
			deg_range=(tx_range+rx_range)/69.0;

			/* Prevent the demand for a really wide coverage
			   from allocating more slots than are available
			   in memory. */

			switch (MAXSLOTS)
			{
				case 2: deg_limit=0.25;
					break;

				case 4: deg_limit=0.5;
					break;

				case 9: deg_limit=1.0;
					break;

				case 16: deg_limit=2.0;
					break;

				case 25: deg_limit=3.0;
			}

			if (deg_range>deg_limit)
				deg_range=deg_limit;

			north_min=(unsigned char)floor(tx_site[z].lat-deg_range);
			north_max=(unsigned char)floor(tx_site[z].lat+deg_range);
			west_min=(unsigned char)floor(tx_site[z].lon-deg_range);
			west_max=(unsigned char)floor(tx_site[z].lon+deg_range);

			if (min_lat==0)
				min_lat=north_min;

			else if (north_min<min_lat)
				min_lat=north_min;

			if (min_lon==0)
				min_lon=west_min;

			else if (west_min<min_lon)
				min_lon=west_min;

			if (north_max>max_lat)
				max_lat=north_max;

			if (west_max>max_lon)
				max_lon=west_max;
		}

		if (min_lat!=0 && min_lon!=0 && max_lat!=0 && max_lon!=0)
		{
			for (y=min_lon; y<=max_lon; y++)
				for (x=min_lat; x<=max_lat; x++)
				{
					sprintf(string,"%u:%u:%u:%u",x, x+1, y, y+1);
					LoadSDF(string);
				}
		}
	}

	if (mapfile[0])
		map=1;

	if (coverage)
	{
		for (x=0; x<txsites; x++)
		{
			PlotCoverage(tx_site[x],altitude);
			PlaceMarker(tx_site[x]);

			if (report!='N')
				SiteReport(tx_site[x]);
		}

		map=1;
	}
	
	else
	{
		PlaceMarker(rx_site);

		for (x=0; x<txsites; x++)
		{
			PlaceMarker(tx_site[x]);

			if (report=='y')
			{
				switch (x)
				{
					case 0:
						PlotPath(tx_site[x],rx_site,1);
						break;

					case 1:
						PlotPath(tx_site[x],rx_site,8);
						break;

					case 2:
						PlotPath(tx_site[x],rx_site,16);
						break;

					case 3:
						PlotPath(tx_site[x],rx_site,32);
				}
			}

			if (report!='N')
				ObstructionReport(tx_site[x],rx_site,report);
		}
	}

	if (map)
	{
		if (bfs)
		{
			for (x=0; x<bfs; x++)
				LoadBoundaries(boundary_file[x]);
		}

		if (cities)
		{
			for (x=0; x<cities; x++)
				LoadCities(city_file[x]);
		}
				
		WritePPM(mapfile);
	}

	if (terrain_plot)
	{
		if (txsites>1)
		{
			for (x=0; terrain_file[x]!='.' && terrain_file[x]!=0 && x<80; x++);

			if (terrain_file[x]=='.')  /* extension */
			{
				ext[0]='.';
				for (y=1, z=x, x++; terrain_file[x]!=0 && x<253 && y<14; x++, y++)
					ext[y]=terrain_file[x];

				ext[y]=0;
				terrain_file[z]=0;
			}

			else
			{
				ext[0]=0;  /* No extension */
				terrain_file[x]=0;
			}

			for (count=0; count<txsites; count++)
			{
				sprintf(string,"%s-%c%s%c",terrain_file,'1'+count,ext,0);
				GraphTerrain(tx_site[count],rx_site,string);
			}
		}

		else
			GraphTerrain(tx_site[0],rx_site,terrain_file);
	}

	if (elevation_plot)
	{
		if (txsites>1)
		{
			for (x=0; elevation_file[x]!='.' && elevation_file[x]!=0 && x<80; x++);

			if (elevation_file[x]=='.')  /* extension */
			{
				ext[0]='.';
				for (y=1, z=x, x++; elevation_file[x]!=0 && x<253 && y<14; x++, y++)
					ext[y]=elevation_file[x];

				ext[y]=0;
				elevation_file[z]=0;
			}

			else
			{
				ext[0]=0;  /* No extension */
				elevation_file[x]=0;
			}

			for (count=0; count<txsites; count++)
			{
				sprintf(string,"%s-%c%s%c",elevation_file,'1'+count,ext,0);
				GraphElevation(tx_site[count],rx_site,string);
			}
		}

		else
			GraphElevation(tx_site[0],rx_site,elevation_file);
	}
	exit(0);
}
